/* ZoiteChat
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifndef ZOITECHAT_XTEXT_COLOR_H
#define ZOITECHAT_XTEXT_COLOR_H

typedef struct
{
	double red;
	double green;
	double blue;
	double alpha;
} XTextColor;

#endif
